package mvd.jester;

import java.lang.reflect.Constructor;
import java.time.LocalTime;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import com.google.common.collect.Table;
import mvd.jester.model.SystemSetup;
import mvd.jester.priority.PriorityManager;
import mvd.jester.simulator.AbstractSimulator;
import mvd.jester.tests.AbstractTest;
import mvd.jester.utils.Logger;

/**
 * TestEnvironment
 */
public class TestEnvironment {

    private final long numberOfTaskSets;
    private final SystemSetup systemSetup;
    private final SystemSetup.Builder builder;
    private final Set<Constructor<? extends AbstractTest>> abstractTests;
    private final Set<Constructor<? extends AbstractSimulator>> abstractSimulators;
    private final Set<PriorityManager> schedulingAlgorithms;

    public TestEnvironment(SystemSetup.Builder builder, long numberOfTaskSets) {
        this.numberOfTaskSets = numberOfTaskSets;
        this.abstractTests = new HashSet<>();
        this.abstractSimulators = new HashSet<>();
        this.schedulingAlgorithms = new HashSet<>();

        this.builder = builder;
        this.systemSetup = builder.build();
    }

    public TestEnvironment registerSchedulingAlgorithm(PriorityManager priorityManager) {
        schedulingAlgorithms.add(priorityManager);

        return this;
    }

    public TestEnvironment registerTest(Class<? extends AbstractTest> abstractTest) {
        try {
            abstractTests.add(abstractTest.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract test!");
        }

        return this;
    }

    public TestEnvironment registerSimulator(Class<? extends AbstractSimulator> abstractSimulator) {
        try {
            abstractSimulators.add(abstractSimulator.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract simulator!");
        }

        return this;
    }

    public TestEnvironment registerTestPair(Class<? extends AbstractTest> abstractTest,
            Class<? extends AbstractSimulator> abstractSimulator) {
        try {
            abstractTests.add(abstractTest.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract test!");
        }
        try {
            abstractSimulators.add(abstractSimulator.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor for abstract simulator!");
        }

        return this;
    }

    public void runTests() {
        Set<AbstractTest> abstractTestInstances = new HashSet<>();
        Set<AbstractSimulator> abstractSimulatorInstances = new HashSet<>();

        Map<Long, Map<AbstractTest, Long>> testResults = new HashMap<>();
        Map<Long, Map<AbstractSimulator, Long>> simulatorResults = new HashMap<>();
        // Map<Long, Table<PriorityManager, AbstractTest, Long>> testResults = new HashMap<>();
        // Map<Long, Table<PriorityManager, AbstractSimulator, Long>> simulatorResults =
        // new HashMap<>();

        for (Constructor<? extends AbstractTest> c : abstractTests) {
            try {
                abstractTestInstances.add(c.newInstance(this.systemSetup));
            } catch (Exception e) {
                System.out.println("Could not instantiate object of AbstractTest!");
            }
        }

        for (Constructor<? extends AbstractSimulator> c : abstractSimulators) {
            try {
                abstractSimulatorInstances.add(c.newInstance(this.systemSetup));
            } catch (Exception e) {
                System.out.println("Could not instantiate object of AbstractSimulator!");
            }
        }

        long checkedTasksets = 0;

        while (checkedTasksets < numberOfTaskSets) {
            builder.rebuild(this.systemSetup);

            double utilization = this.systemSetup.getUtilization();

            while (utilization <= this.systemSetup.getNumberOfProcessors()
                    && checkedTasksets < numberOfTaskSets) {
                checkedTasksets++;

                if (checkedTasksets % 100 == 0) {
                    System.out.println(checkedTasksets + " bereits geschafft");
                }

                long roundedUtilization = (long) (utilization * 10);

                for (PriorityManager priorityManager : schedulingAlgorithms) {
                    for (AbstractTest test : abstractTestInstances) {
                        if (priorityManager.hasTest(test)
                                && test.runSchedulabilityCheck(priorityManager)) {
                            testResults
                                    .computeIfAbsent(roundedUtilization,
                                            k -> new HashMap<AbstractTest, Long>())
                                    .compute(test, (k, v) -> (v == null) ? 1 : v + 1);
                        }
                    }

                    for (AbstractSimulator simulator : abstractSimulatorInstances) {
                        if (priorityManager.hasSimulator(simulator)
                                && simulator.runSimulation(priorityManager)) {
                            simulatorResults
                                    .computeIfAbsent(roundedUtilization,
                                            k -> new HashMap<AbstractSimulator, Long>())
                                    .compute(simulator, (k, v) -> (v == null) ? 1 : v + 1);
                        }
                    }

                }
                builder.addTask(systemSetup);

                utilization = this.systemSetup.getUtilization();
            }
        }

        logTestResults(abstractTestInstances, testResults);
        logSimulationResults(abstractSimulatorInstances, simulatorResults);
    }

    private void logTestResults(Set<AbstractTest> testCases,
            Map<Long, Map<AbstractTest, Long>> results) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/results_" + systemSetup.getNumberOfProcessors() + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");
        String firstLine = new String("Utilization");

        for (AbstractTest t : testCases) {
            firstLine = firstLine + "\t" + t.getName();
        }

        log.log(firstLine);

        for (Long util : results.keySet()) {
            String line = String.valueOf((double) util / 10);
            Map<AbstractTest, Long> tests = results.get(util);

            for (AbstractTest t : testCases) {
                if (tests.containsKey(t)) {
                    line += "\t" + tests.get(t);
                } else {
                    line += "\t" + "0";
                }
            }
            log.log(line);
        }

        log.finalize();
    }

    private void logSimulationResults(Set<AbstractSimulator> testCases,
            Map<Long, Map<AbstractSimulator, Long>> results) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/results_sim_" + systemSetup.getNumberOfProcessors() + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");
        String firstLine = new String("Utilization");

        for (AbstractSimulator t : testCases) {
            firstLine = firstLine + "\t" + t.getName();
        }

        log.log(firstLine);

        for (Long util : results.keySet()) {
            String line = String.valueOf((double) util / 10);
            Map<AbstractSimulator, Long> tests = results.get(util);

            for (AbstractSimulator t : testCases) {
                if (tests.containsKey(t)) {
                    line += "\t" + tests.get(t);
                } else {
                    line += "\t" + "0";
                }
            }
            log.log(line);
        }

        log.finalize();
    }
}
